/*
 * Card.java
 *
 * Contains the Card class, and the Card.Rank and Card.Suit enumerations,
 * implementing standard playing cards.
 *
 * Modified 2009-10-11 by Nicholas Bennett.
 */

package org.nm.challenge.examples.cards;

import java.util.Arrays;

/**
 * <p>This class implements standard playing cards. An instance of this class
 * corresponds to a single card, and the {@link #newDeck() newDeck()}
 * static method can be used to get an entire deck of cards in an array.</p>
 *
 * <p>Instances of this class are immutable (and consequently, thread-safe). For
 * many applications, there will be no need to create more than 52 instances of
 * this class (though it might be convenient to create more, for special cases).
 * The <code>newDeck()</code> method takes advantage of this, by returning the
 * same <code>Card</code> instances each time, but in a new array.</p>
 *
 * <p>This class is based on the illustration of enumerations at <a
 * href="http://java.sun.com/j2se/1.5.0/docs/guide/language/enums.html">
 * http://java.sun.com/j2se/1.5.0/docs/guide/language/enums.html</a>.</p>
 *
 * @author Nicholas Bennett
 * @version 1.0, 2009-10-11
 */
public class Card implements Comparable {

    private static final String TO_STRING_PATTERN = "%s%s";

    private static Card[] prototypeDeck;

    private Rank rank;
    private Suit suit;

    static {
        prototypeDeck = new Card[Rank.values().length * Suit.values().length];
        int position = 0;
        for (Suit suit : Suit.values()) {
            for (Rank rank : Rank.values()) {
                prototypeDeck[position++] = new Card(rank, suit);
            }
        }
    }

    /**
     * Constructs a new instance, using the specified {@link Rank Rank} and
     * {@link Suit Suit}.
     *
     * @param rank      <code>Rank</code>
     * @param suit      <code>Suit</code>
     */
    public Card(Rank rank, Suit suit) {
        this.rank = rank;
        this.suit = suit;
    }

    /**
     * Returns the {@link Rank Rank} of this <code>Card</code> instance.
     * 
     * @return          <code>Rank</code>
     */
    public Rank getRank() {
        return this.rank;
    }

    /**
     * Returns the {@link Suit Suit} of this <code>Card</code> instance.
     *
     * @return          <code>Suit</code>
     */
    public Suit getSuit() {
        return this.suit;
    }

    /**
     * Returns an unshuffled deck of cards as an array.
     *
     * @return          <code>Card[]</code>
     */
    public static Card[] newDeck() {
        return Arrays.copyOf(prototypeDeck, prototypeDeck.length);
    }

    /**
     * Returns a <code>string</code> representation of the <code>Card</code>,
     * consisting of the rank number or letter, followed by the suit symbol.
     *
     * @return          <code>String</code>
     */@Override
    public String toString() {
        return String.format(TO_STRING_PATTERN, this.rank, this.suit);
    }

    /**
     * Compares one <code>Card</code> instance to another, for the purpose of
     * sorting. The comparison is in order of rank, then suit.
     * 
     * @param o         <code>Object</code> this instance should be compared to.
     * @return          <code>int</code>
     */
    public int compareTo(Object o) {
        int result = 0;
        if (o instanceof Card) {
            Card other = (Card) o;
            result = getRank().compareTo(other.getRank());
            if (result == 0) {
                result = getSuit().compareTo(other.getSuit());
            }
        }
        else {
            result = toString().compareTo(o.toString());
        }
        return result;
    }

    /**
     * This is an enumeration of all 13 ranks in a standard deck of playing
     * cards.
     */
    public enum Rank {

        TWO("2"),
        THREE("3"),
        FOUR("4"),
        FIVE("5"),
        SIX("6"),
        SEVEN("7"),
        EIGHT("8"),
        NINE("9"),
        TEN("10"),
        JACK("J"),
        QUEEN("Q"),
        KING("K"),
        ACE("A");

        private String display;

        Rank(String display) {
            this.display = display;
        }

        @Override
        public String toString() {
            return display;
        }
        
    }

    /**
     * This is an enumeration of the 4 suits in a standard deck of playing
     * cards.
     */
    public enum Suit {

        CLUBS((char) 0x2663),
        DIAMONDS((char) 0x2662),
        HEARTS((char) 0x2661),
        SPADES((char) 0x2660);

        private char display;

        Suit(char display) {
            this.display = display;
        }

        @Override
        public String toString() {
            return Character.toString(display);
        }

    }

}
